API_URL = "https://m.airzonecloud.com/api/v1"

MODES_CONVERTER = {
    "0": {
        "name": "stop",
        "generic": "stop",
        "description": "Stop mode",
        "range_key_prefix": "range_sp_stop_air_",
        "setpoint_key": "setpoint_air_stop",
    },
    "1": {
        "name": "auto",
        "generic": "auto",
        "description": "Automatic mode",
        "range_key_prefix": "range_sp_auto_air_",
        "setpoint_key": "setpoint_air_auto",
    },
    "2": {
        "name": "cooling",
        "generic": "cooling",
        "description": "Cooling mode",
        "range_key_prefix": "range_sp_cool_air_",
        "setpoint_key": "setpoint_air_cool",
    },
    "3": {
        "name": "heating",
        "generic": "heating",
        "description": "Heating mode",
        "range_key_prefix": "range_sp_hot_air_",
        "setpoint_key": "setpoint_air_heat",
    },
    "4": {
        "name": "ventilation",
        "generic": "ventilation",
        "description": "Ventilation mode",
        "range_key_prefix": "range_sp_vent_air_",
        "setpoint_key": "setpoint_air_vent",
    },
    "5": {
        "name": "dehumidify",
        "generic": "dehumidify",
        "description": "Dehumidifier / Dry mode",
        "range_key_prefix": "range_sp_dry_air_",
        "setpoint_key": "setpoint_air_dry",
    },
    "6": {
        "name": "emergency-heating",
        "generic": "emergency",
        "description": "Emergency heat mode",
        "range_key_prefix": "range_sp_emerheat_air_",
        "setpoint_key": "setpoint_air_emerheat",
    },
    "7": {
        "name": "air-heating",
        "generic": "heating",
        "description": "Heat air mode (only compatible systems)",
        "range_key_prefix": "range_sp_hot_air_",
        "setpoint_key": "setpoint_air_heat",
    },
    "8": {
        "name": "radiant-heating",
        "generic": "heating",
        "description": "Heat radiant mode (only compatible systems)",
        "range_key_prefix": "range_sp_hot_air_",
        "setpoint_key": "setpoint_air_heat",
    },
    "9": {
        "name": "combined-heating",
        "generic": "heating",
        "description": "Heat combined mode (only compatible systems)",
        "range_key_prefix": "range_sp_hot_air_",
        "setpoint_key": "setpoint_air_heat",
    },
    "10": {
        "name": "air-cooling",
        "generic": "cooling",
        "description": "Cooling air mode (only compatible systems)",
        "range_key_prefix": "range_sp_cool_air_",
        "setpoint_key": "setpoint_air_cool",
    },
    "11": {
        "name": "radiant-cooling",
        "generic": "cooling",
        "description": "Cooling radiant mode (only compatible systems)",
        "range_key_prefix": "range_sp_cool_air_",
        "setpoint_key": "setpoint_air_cool",
    },
    "12": {
        "name": "combined-cooling",
        "generic": "cooling",
        "description": "Cooling combined mode (only compatible systems)",
        "range_key_prefix": "range_sp_cool_air_",
        "setpoint_key": "setpoint_air_cool",
    },
}



import logging
import time
from typing import Union
from . import AirzoneCloud, Group
from .constants import MODES_CONVERTER

_LOGGER = logging.getLogger(__name__)


class Device:
    """Manage a AirzoneCloud device (thermostat)"""

    _api: "AirzoneCloud" = None
    _group: "Group" = None
    _data: dict = {}
    _state: dict = {}

    def __init__(self, api: "AirzoneCloud", group: "Group", data: dict) -> None:
        self._api = api
        self._group = group
        self._data = data

        # load state
        self.refresh()

        # log
        _LOGGER.info("Init {}".format(self.str_verbose))
        _LOGGER.debug(data)

    def __str__(self) -> str:
        return "Device(name={}, is_connected={}, is_on={}, mode={}, current_temp={}, target_temp={})".format(
            self.name,
            self.is_connected,
            self.is_on,
            self.mode,
            self.current_temperature,
            self.target_temperature,
        )

    @property
    def str_verbose(self) -> str:
        """More verbose description of current device"""
        return "Device(name={}, is_connected={}, is_on={}, mode={}, current_temp={}, target_temp={}, id={}, ws_id={})".format(
            self.name,
            self.is_connected,
            self.is_on,
            self.mode,
            self.current_temperature,
            self.target_temperature,
            self.id,
            self.ws_id,
        )

    @property
    def all_properties(self) -> dict:
        """Return all group properties values"""
        result = {}
        for prop in [
            "id",
            "name",
            "type",
            "ws_id",
            "system_number",
            "zone_number",
            "is_connected",
            "is_on",
            "mode_id",
            "mode",
            "mode_generic",
            "mode_description",
            "modes_availables",
            "modes_availables_generics",
            "current_humidity",
            "current_temperature",
            "target_temperature",
            "min_temperature",
            "max_temperature",
            "step_temperature",
        ]:
            result[prop] = getattr(self, prop)
        return result

    #
    # getters
    #

    @property
    def id(self) -> str:
        """Return device id"""
        return self._data.get("device_id")

    @property
    def name(self) -> str:
        """Return device name"""
        return self._data.get("name")

    @property
    def type(self) -> str:
        """Return device type (az_zone┃aidoo)"""
        return self._data.get("type")

    @property
    def ws_id(self) -> str:
        """Return device webserver id (mac address)"""
        return self._data.get("ws_id")

    @property
    def system_number(self) -> str:
        """Return device system_number"""
        return self._data.get("meta", {}).get("system_number")

    @property
    def zone_number(self) -> str:
        """Return device zone_number"""
        return self._data.get("meta", {}).get("zone_number")

    @property
    def is_connected(self) -> bool:
        """Return if the device is online (True) or offline (False)"""
        return self._state.get("isConnected", False)

    @property
    def is_on(self) -> bool:
        """Return True if the device is on"""
        return self._state.get("power", False)

    @property
    def is_master(self) -> bool:
        """Return True if the device is a master thermostat (allowed to update the mode of all devices)"""
        return len(self.modes_availables_ids) > 0

    @property
    def mode_id(self) -> int:
        """Return device current id mode (0┃1┃2┃3┃4┃5┃6┃7┃8┃9┃10┃11┃12)"""
        return self._state.get("mode", 0)

    @property
    def mode(self) -> str:
        """Return device current mode name (stop | auto | cooling | heating | ventilation | dehumidify | emergency-heating | air-heating | radiant-heating | combined-heating | air-cooling | radiant-cooling | combined-cooling)"""
        return MODES_CONVERTER.get(str(self.mode_id), {}).get("name")

    @property
    def mode_generic(self) -> str:
        """Return device current generic mode (stop | auto | cooling | heating | ventilation | dehumidify | emergency)"""
        return MODES_CONVERTER.get(str(self.mode_id), {}).get("generic")

    @property
    def mode_description(self) -> str:
        """Return device current mode description (pretty name to display)"""
        return MODES_CONVERTER.get(str(self.mode_id), {}).get("description")

    @property
    def modes_availables_ids(self) -> "list[int]":
        """Return device availables modes list ([0┃1┃2┃3┃4┃5┃6┃7┃8┃9┃10┃11┃12, ...])"""
        return self._state.get("mode_available", [])

    @property
    def modes_availables(self) -> "list[str]":
        """Return device availables modes names list ([stop | auto | cooling | heating | ventilation | dehumidify | emergency-heating | air-heating | radiant-heating | combined-heating | air-cooling | radiant-cooling | combined-cooling, ...])"""
        return [
            MODES_CONVERTER.get(str(mode_id), {}).get("name")
            for mode_id in self.modes_availables_ids
        ]

    @property
    def modes_availables_generics(self) -> "list[str]":
        """Return device availables modes generics list ([stop | auto | cooling | heating | ventilation | dehumidify | emergency, ...])"""
        return list(
            set(
                [
                    MODES_CONVERTER.get(str(mode_id), {}).get("generic")
                    for mode_id in self.modes_availables_ids
                ]
            )
        )

    @property
    def current_temperature(self) -> float:
        """Return device current temperature in °C"""
        return float(self._state.get("local_temp", {}).get("celsius", 0))

    @property
    def current_humidity(self) -> int:
        """Return device current humidity in percentage (0-100)"""
        return int(self._state.get("humidity", 0))

    @property
    def target_temperature(self) -> float:
        """Return device target temperature for current mode"""
        key = MODES_CONVERTER.get(str(self.mode_id), {}).get("setpoint_key")
        return float(self._state.get(key, {}).get("celsius", 0))

    @property
    def min_temperature(self) -> float:
        """Return device minimal temperature for current mode"""
        key = MODES_CONVERTER.get(str(self.mode_id), {}).get("range_key_prefix") + "min"
        return float(self._state.get(key, {}).get("celsius", 0))

    @property
    def max_temperature(self) -> float:
        """Return device maximal temperature for current mode"""
        key = MODES_CONVERTER.get(str(self.mode_id), {}).get("range_key_prefix") + "max"
        return float(self._state.get(key, {}).get("celsius", 0))

    @property
    def step_temperature(self) -> float:
        """Return device step temperature (minimum increase/decrease step)"""
        return float(self._state.get("step", {}).get("celsius", 0.5))

    #
    # setters
    #

    def turn_on(self, auto_refresh: bool = True, delay_refresh: int = 1) -> "Device":
        """Turn device on"""
        _LOGGER.info("call turn_on() on {}".format(self.str_verbose))

        self._set("power", True)

        if auto_refresh:
            time.sleep(delay_refresh)  # wait data refresh by airzone
            self.refresh()

        return self

    def turn_off(self, auto_refresh: bool = True, delay_refresh: int = 1) -> "Device":
        """Turn device off"""
        _LOGGER.info("call turn_off() on {}".format(self.str_verbose))

        self._set("power", False)

        if auto_refresh:
            time.sleep(delay_refresh)  # wait data refresh by airzone
            self.refresh()

        return self

    def set_temperature(
        self, temperature: float, auto_refresh: bool = True, delay_refresh: int = 1
    ) -> "Device":
        """Set target_temperature for current device (degrees celsius)"""
        _LOGGER.info(
            "call set_temperature({}) on {}".format(temperature, self.str_verbose)
        )
        if self.min_temperature is not None and temperature < self.min_temperature:
            temperature = self.min_temperature
        if self.max_temperature is not None and temperature > self.max_temperature:
            temperature = self.max_temperature

        self._set("setpoint", temperature)

        if auto_refresh:
            time.sleep(delay_refresh)  # wait data refresh by airzone
            self.refresh()

        return self

    def set_mode(
        self, mode_name: str, auto_refresh: bool = True, delay_refresh: int = 1
    ) -> "Device":
        """Set mode of the device"""
        _LOGGER.info("call set_mode({}) on {}".format(mode_name, self.str_verbose))

        # search mode id
        mode_id_found = None
        for mode_id, mode in MODES_CONVERTER.items():
            if mode["name"] == mode_name:
                mode_id_found = int(mode_id)
                break
        if mode_id_found is None:
            raise ValueError(
                'mode name "{}" not found for {}'.format(mode_name, self.str_verbose)
            )

        if mode_id_found not in self.modes_availables_ids:
            if len(self.modes_availables_ids) == 0:
                raise ValueError(
                    'mode name "{}" (id: {}) not availables for {} : only master thermostat device can set the mode'.format(
                        mode_name,
                        mode_id_found,
                        self.str_verbose,
                        self.modes_availables,
                    )
                )

            raise ValueError(
                'mode name "{}" (id: {}) not availables for {}. Allowed values: {}'.format(
                    mode_name, mode_id_found, self.str_verbose, self.modes_availables
                )
            )

        self._set("mode", mode_id_found)

        if auto_refresh:
            time.sleep(delay_refresh)  # wait data refresh by airzone
            self.refresh()

        return self

    #
    # parent group
    #

    @property
    def group(self) -> Group:
        """Get parent group"""
        return self._group

    #
    # Refresh
    #

    def refresh(self) -> "Device":
        """Refresh current device states"""
        _LOGGER.debug("call refresh() on {}".format(self.str_verbose))
        self._state = self._api._api_get_device_state(
            self.id, self.group.installation.id
        )
        _LOGGER.debug(self._state)
        return self

    #
    # private
    #

    def _set(self, param: str, value: Union[str, int, float, bool]) -> "Device":
        """Execute a command to the current device (power, mode, setpoint, ...)"""
        _LOGGER.debug("call _set({}, {}) on {}".format(param, value, self.str_verbose))
        self._api._api_patch_device(
            self.id, self.group.installation.id, param, value, {"units": 0}
        )
        return self

    def _set_data_refreshed(self, data: dict) -> "Device":
        """Set data refreshed (called by parent AirzoneCloud on refresh_devices())"""
        self._data = data
        _LOGGER.info("Data refreshed for {}".format(self.str_verbose))
        return self


#
# device raw data example
#

# {
#     "device_id": "60f5cb9...",
#     "meta": {
#         "system_number": 1,
#         "zone_number": 6
#     },
#     "type": "az_zone",
#     "ws_id": "AA:BB:CC:DD:EE:FF",
#     "name": "Salon"
# }

#
# device raw state example
#

# {
#   "active": null,
#   "aq_active": null,
#   "aq_mode_conf": null,
#   "aq_mode_values": [],
#   "aq_quality": null,
#   "aqpm1_0": null,
#   "aqpm2_5": null,
#   "aqpm10": null,
#   "auto_mode": null,
#   "connection_date": "2021-11-17T08:44:04.000Z",
#   "disconnection_date": "2021-11-16T06:24:11.499Z",
#   "eco_conf": "off",
#   "eco_values": [
#     "off",
#     "manual",
#     "a",
#     "a_p",
#     "a_pp"
#   ],
#   "humidity": 48,
#   "isConnected": true,
#   "local_temp": {
#     "celsius": 20.7,
#     "fah": 69
#   },
#   "mode": 3,
#   "mode_available": [
#     2,
#     3,
#     4,
#     5,
#     0
#   ],
#   "name": "Salon",
#   "power": true,
#   "range_sp_cool_air_max": {
#     "fah": 86,
#     "celsius": 30
#   },
#   "range_sp_cool_air_min": {
#     "celsius": 18,
#     "fah": 64
#   },
#   "range_sp_dry_air_max": {
#     "fah": 86,
#     "celsius": 30
#   },
#   "range_sp_dry_air_min": {
#     "celsius": 18,
#     "fah": 64
#   },
#   "range_sp_emerheat_air_max": {
#     "celsius": 30,
#     "fah": 86
#   },
#   "range_sp_emerheat_air_min": {
#     "fah": 59,
#     "celsius": 15
#   },
#   "range_sp_hot_air_max": {
#     "celsius": 30,
#     "fah": 86
#   },
#   "range_sp_hot_air_min": {
#     "fah": 59,
#     "celsius": 15
#   },
#   "range_sp_stop_air_max": {
#     "fah": 86,
#     "celsius": 30
#   },
#   "range_sp_stop_air_min": {
#     "fah": 59,
#     "celsius": 15
#   },
#   "range_sp_vent_air_max": {
#     "fah": 86,
#     "celsius": 30
#   },
#   "range_sp_vent_air_min": {
#     "fah": 59,
#     "celsius": 15
#   },
#   "setpoint_air_heat": {
#     "celsius": 20,
#     "fah": 68
#   },
#   "setpoint_air_stop": {
#     "celsius": 20,
#     "fah": 68
#   },
#   "sleep": 0,
#   "sleep_values": [
#     0,
#     30,
#     60,
#     90
#   ],
#   "speed_conf": null,
#   "speed_type": null,
#   "speed_values": [],
#   "step": {
#     "fah": 1,
#     "celsius": 0.5
#   },
#   "usermode_conf": null,
#   "warnings": [],
#   "zone_sched_available": false
# }



import logging
import time
from typing import Union
from . import AirzoneCloud, Installation
from .constants import MODES_CONVERTER
from .Device import Device

_LOGGER = logging.getLogger(__name__)


class Group:
    """Manage a AirzoneCloud group"""

    _api: AirzoneCloud = None
    _installation: Installation = None
    _data: dict = {}
    _devices: "list[Device]" = []

    def __init__(
        self, api: AirzoneCloud, installation: Installation, data: dict
    ) -> None:
        self._api = api
        self._installation = installation
        self._data = data

        # log
        _LOGGER.info("Init {}".format(self.str_verbose))
        _LOGGER.debug(data)

        # load all devices
        self._load_devices()

    def __str__(self) -> str:
        return "Group(name={}, installation={})".format(
            self.name, self._installation.name
        )

    @property
    def str_verbose(self) -> str:
        """More verbose description of current group"""
        return "Group(name={}, installation={}, id={})".format(
            self.name, self._installation.name, self.id
        )

    @property
    def all_properties(self) -> dict:
        """Return all group properties values"""
        result = {}
        for prop in [
            "id",
            "name",
            "is_on",
            "mode_id",
            "mode",
            "mode_generic",
            "mode_description",
            "modes_availables",
            "modes_availables_generics",
        ]:
            result[prop] = getattr(self, prop)
        return result

    #
    # getters
    #

    @property
    def id(self) -> str:
        """Return group id"""
        return self._data.get("group_id")

    @property
    def name(self) -> str:
        """Return group name"""
        return self._data.get("name")

    @property
    def is_on(self) -> bool:
        """Return True if at least one device is on in the group"""
        for device in self.devices:
            if device.is_on:
                return True
        return False

    @property
    def mode_id(self) -> int:
        """Return group current id mode (0┃1┃2┃3┃4┃5┃6┃7┃8┃9┃10┃11┃12)"""
        return self.master_device.mode_id

    @property
    def mode(self) -> str:
        """Return group current mode name (stop | auto | cooling | heating | ventilation | dehumidify | emergency-heating | air-heating | radiant-heating | combined-heating | air-cooling | radiant-cooling | combined-cooling)"""
        return self.master_device.mode

    @property
    def mode_generic(self) -> str:
        """Return group current generic mode (stop | auto | cooling | heating | ventilation | dehumidify | emergency)"""
        return self.master_device.mode_generic

    @property
    def mode_description(self) -> str:
        """Return group current mode description (pretty name to display)"""
        return self.master_device.mode_description

    @property
    def modes_availables_ids(self) -> "list[int]":
        """Return group availables modes list ([0┃1┃2┃3┃4┃5┃6┃7┃8┃9┃10┃11┃12, ...])"""
        return self.master_device.modes_availables_ids

    @property
    def modes_availables(self) -> "list[str]":
        """Return group availables modes names list ([stop | auto | cooling | heating | ventilation | dehumidify | emergency-heating | air-heating | radiant-heating | combined-heating | air-cooling | radiant-cooling | combined-cooling, ...])"""
        return self.master_device.modes_availables

    @property
    def modes_availables_generics(self) -> "list[str]":
        """Return group availables modes generics list ([stop | auto | cooling | heating | ventilation | dehumidify | emergency, ...])"""
        return self.master_device.modes_availables_generics

    #
    # setters
    #

    def turn_on(self, auto_refresh: bool = True, delay_refresh: int = 1) -> "Group":
        """Turn on all devices in the group"""
        _LOGGER.info("call turn_on() on {}".format(self.str_verbose))

        for device in self.devices:
            device.turn_on(auto_refresh=False)

        if auto_refresh:
            time.sleep(delay_refresh)  # wait data refresh by airzone
            self.refresh_devices()

        return self

    def turn_off(self, auto_refresh: bool = True, delay_refresh: int = 1) -> "Group":
        """Turn off all devices in the group"""
        _LOGGER.info("call turn_off() on {}".format(self.str_verbose))

        for device in self.devices:
            device.turn_off(auto_refresh=False)

        if auto_refresh:
            time.sleep(delay_refresh)  # wait data refresh by airzone
            self.refresh_devices()

        return self

    def set_temperature(
        self, temperature: float, auto_refresh: bool = True, delay_refresh: int = 1
    ) -> "Group":
        """Set target_temperature for current all devices in the group (in degrees celsius)"""
        _LOGGER.info(
            "call set_temperature({}) on {}".format(temperature, self.str_verbose)
        )

        for device in self.devices:
            device.set_temperature(temperature=temperature, auto_refresh=False)

        if auto_refresh:
            time.sleep(delay_refresh)  # wait data refresh by airzone
            self.refresh_devices()

        return self

    def set_mode(
        self, mode_name: str, auto_refresh: bool = True, delay_refresh: int = 1
    ) -> "Group":
        """Set mode of the all devices in the group"""
        _LOGGER.info("call set_mode({}) on {}".format(mode_name, self.str_verbose))

        self.master_device.set_mode(
            mode_name=mode_name, auto_refresh=auto_refresh, delay_refresh=delay_refresh
        )

        return self

    #
    # parent installation
    #

    @property
    def installation(self) -> Installation:
        """Get parent installation"""
        return self._installation

    #
    # children
    #

    @property
    def devices(self) -> "list[Device]":
        """Return all devices in this group"""
        return self._devices

    @property
    def master_device(self) -> "Device":
        """Return master device in this group (only device allowed to change mode)"""
        for device in self.devices:
            if device.is_master:
                return device
        raise Exception(
            "Cannot find master device in group {}".format(self.str_verbose)
        )

    #
    # Refresh
    #

    def refresh_devices(self) -> "Group":
        """Refresh all devices of this group"""
        for device in self.devices:
            device.refresh()
        return self

    #
    # private
    #

    def _load_devices(self) -> "list[Device]":
        """Load all devices for this group"""
        previous_devices = self._devices
        self._devices = []
        for device_data in self._data.get("devices", []):
            # skip fake system device
            if device_data.get("type") != "az_zone":
                continue
            device = None
            # search device in previous_devices (if where are refreshing devices)
            for previous_device in previous_devices:
                if previous_device.id == device_data.get("device_id"):
                    device = previous_device
                    # update data
                    device._set_data_refreshed(device_data)
                    break
            # device not found => instance new device
            if device is None:
                device = Device(self._api, self, device_data)
            self._devices.append(device)
        return self._devices

    def _set(self, param: str, value: Union[str, int, float, bool]) -> "Group":
        """Execute a command to the current device (power, mode, setpoint, ...)"""
        _LOGGER.debug("call _set({}, {}) on {}".format(param, value, self.str_verbose))
        self._api._api_put_group(
            self.id, self.group.installation.id, param, value, {"units": 0}
        )
        return self

    def _set_data_refreshed(self, data: dict) -> "Group":
        """Set data refreshed (called by parent Installation on refresh_groups())"""
        self._data = data
        _LOGGER.info("Data refreshed for {}".format(self.str_verbose))
        return self


#
# group raw data example
#

# {
#     "group_id": "60f5cb99ff517e33f0365733",
#     "name": "Système 1",
#     "devices": [
#         {
#             "device_id": "60f5cb...",
#             "meta": {"system_number": 1},
#             "type": "az_system",
#             "ws_id": "AA:BB:CC:DD:EE:FF",
#         },
#         {
#             "device_id": "60f5cb...",
#             "meta": {"system_number": 1, "zone_number": 6},
#             "type": "az_zone",
#             "ws_id": "AA:BB:CC:DD:EE:FF",
#             "name": "Salon",
#         },
#         {
#             "device_id": "60f5cb991...",
#             "meta": {"system_number": 1, "zone_number": 4},
#             "type": "az_zone",
#             "ws_id": "AA:BB:CC:DD:EE:FF",
#             "name": "Ch parents",
#         },
#         {
#             "device_id": "60f5cb9...",
#             "meta": {"system_number": 1, "zone_number": 5},
#             "type": "az_zone",
#             "ws_id": "AA:BB:CC:DD:EE:FF",
#             "name": "Ch bebe",
#         },
#     ],
# }



#!/usr/bin/python3

import logging
from typing import Any, Union
import requests
import urllib
import urllib.parse

from .Installation import Installation
from .Group import Group
from .Device import Device
from .constants import API_URL

_LOGGER = logging.getLogger(__name__)


class AirzoneCloud:
    """Allow to connect to AirzoneCloud API"""

    _email: str = None
    _password: str = None
    _user_agent: str = "Mozilla/5.0 (Linux; Android 6.0.1; Nexus 7 Build/MOB30X; wv) AppleWebKit/537.26 (KHTML, like Gecko) Version/4.0 Chrome/70.0.3538.110 Safari/537.36"
    _session: requests.Session = None
    _token: str = None
    _installations: "list[Installation]" = []

    def __init__(self, email: str, password: str, user_agent: str = None) -> None:
        """Initialize API connection"""
        self._email = email
        self._password = password
        if user_agent is not None and isinstance(user_agent, str):
            self._user_agent = user_agent

        # init new Session
        self._session = requests.Session()

        # login
        self._login()

        # load installations
        self._load_installations()

    #
    # getters
    #

    @property
    def installations(self) -> "list[Installation]":
        """Get installations list"""
        return self._installations

    @property
    def all_groups(self) -> "list[Group]":
        """Get all groups from all installations"""
        result = []
        for installation in self.installations:
            for group in installation.groups:
                result.append(group)
        return result

    @property
    def all_devices(self) -> "list[Device]":
        """Get all devices from all installations"""
        result = []
        for group in self.installations:
            for device in group.all_devices:
                result.append(device)
        return result

    #
    # Refresh
    #

    def refresh_installations(self) -> "AirzoneCloud":
        """Refresh installations"""
        self._load_installations()
        return self

    #
    # private
    #

    def _login(self) -> str:
        """Login to  AirzoneCloud and return token"""

        try:
            url = "{}/auth/login".format(API_URL)
            login_payload = {"email": self._email, "password": self._password}
            headers = {"User-Agent": self._user_agent}
            response = self._session.post(url, headers=headers, json=login_payload)
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise Exception(
                "Unable to login to AirzoneCloud with the email {} and the given password".format(
                    self._email
                )
            ) from None

        self._token = response.json().get("token")
        if not self._token:
            raise Exception(
                "Unable to login to AirzoneCloud, cannot get token from response : {}".format(
                    response
                )
            )

        _LOGGER.info("Login success as {}".format(self._email))

        return self._token

    def _load_installations(self) -> "list[Installation]":
        """Load all installations for this account"""
        previous_installations = self._installations
        self._installations = []
        try:
            for installation_data in self._api_get_installations_list():
                installation = None
                # search installation in previous_installations (if where are refreshing installations)
                for previous_installation in previous_installations:
                    if previous_installation.id == installation_data.get(
                        "installation_id"
                    ):
                        installation = previous_installation
                        installation._set_data_refreshed(installation_data)
                        break
                # installation not found => instance new installation
                if installation is None:
                    installation = Installation(self, installation_data)
                self._installations.append(installation)
        except RuntimeError:
            raise Exception("Unable to load installations from AirzoneCloud")
        return self._installations

    #
    # API calls
    #

    def _api_get_installations_list(self) -> list:
        """Http GET to load installations relations"""
        _LOGGER.debug("_api_get_installations_list()")
        # TODO manage pagination (10 installations max currently)
        return self._api_get("/installations").get("installations", [])

    def _api_get_installation_groups_list(self, installation_id: str) -> list:
        """Http GET to load groups in a specific installation"""
        _LOGGER.debug(
            "_api_get_installation_groups_list(installation_id={})".format(
                installation_id
            )
        )
        return self._api_get("/installations/{}".format(installation_id)).get(
            "groups", []
        )

    def _api_get_device_state(self, device_id: str, installation_id: str) -> dict:
        """Http GET to load state of a specific device"""
        _LOGGER.debug(
            "_api_get_device_state(device_id={}, installation_id={})".format(
                device_id, installation_id
            )
        )
        return self._api_get(
            "/devices/{}/status".format(device_id),
            {"installation_id": installation_id},
        )

    def _api_get_device_config(
        self, device_id: str, installation_id: str, type: str = "all"
    ) -> dict:
        """Http GET to load config of a specific device"""
        _LOGGER.debug(
            "_api_get_device_config(device_id={}, installation_id={}, type={})".format(
                device_id, installation_id, type
            )
        )
        return self._api_get(
            "/devices/{}/config".format(device_id),
            {"installation_id": installation_id, "type": type},
        )

    def _api_patch_device(
        self,
        device_id: str,
        installation_id: str,
        param: str,
        value: Union[str, int, float, bool],
        opts: dict = {},
    ) -> Any:
        """Http PATCH to change a device parameter (state or config)"""
        _LOGGER.debug(
            "_api_patch_device(device_id={}, installation_id={}, param={}, value={}, opts={})".format(
                device_id, installation_id, param, value, opts
            )
        )
        return self._api_patch(
            "/devices/{}".format(device_id),
            {
                "installation_id": installation_id,
                "param": param,
                "value": value,
                "opts": opts,
            },
        )

    def _api_get(self, api_endpoint: str, params: dict = {}) -> Any:
        """Do a http GET request on an api endpoint"""

        params["format"] = "json"

        return self._api_request(method="GET", api_endpoint=api_endpoint, params=params)

    def _api_post(self, api_endpoint: str, payload: dict = {}) -> Any:
        """Do a http POST request on an api endpoint"""

        headers = {
            "Content-Type": "application/json;charset=UTF-8",
            "Accept": "application/json",
        }

        return self._api_request(
            method="POST", api_endpoint=api_endpoint, headers=headers, json=payload
        )

    def _api_put(self, api_endpoint: str, payload: dict = {}) -> Any:
        """Do a http PUT request on an api endpoint"""

        headers = {
            "Content-Type": "application/json;charset=UTF-8",
            "Accept": "application/json",
        }

        return self._api_request(
            method="PUT", api_endpoint=api_endpoint, headers=headers, json=payload
        )

    def _api_patch(self, api_endpoint: str, payload: dict = {}) -> Any:
        """Do a http PATCH request on an api endpoint"""

        headers = {
            "Content-Type": "application/json;charset=UTF-8",
            "Accept": "application/json",
        }

        return self._api_request(
            method="PATCH", api_endpoint=api_endpoint, headers=headers, json=payload
        )

    def _api_request(
        self,
        method: str,
        api_endpoint: str,
        params: dict = {},
        headers: dict = {},
        json: dict = None,
        autoreconnect: bool = True,
    ) -> Any:
        """Do a http generic request on an api endpoint"""

        # set headers
        headers["Authorization"] = "Bearer {}".format(self._token)
        headers["User-Agent"] = self._user_agent

        # generate url
        url = "{}{}/?{}".format(API_URL, api_endpoint, urllib.parse.urlencode(params))

        # make call
        call = self._session.request(method=method, url=url, headers=headers, json=json)

        if call.status_code == 401 and autoreconnect:  # unauthorized error
            # log
            _LOGGER.info(
                "Get unauthorized error (token expired ?), trying to reconnect..."
            )

            # try to reconnect
            self._login()

            # retry get without autoreconnect (to avoid infinite loop)
            return self._api_request(
                method=method,
                api_endpoint=api_endpoint,
                params=params,
                headers=headers,
                json=json,
                autoreconnect=False,
            )

        # raise other error if needed
        try:
            call.raise_for_status()
        except requests.exceptions.HTTPError as err:
            _LOGGER.error(call.text)
            raise err

        # decode json only if response is not empty
        if len(call.text):
            return call.json()

        return None

